// ======================================================================
//
// ObjectList.cpp
// jeff grills
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "First.h"
#include "ObjectList.h"

#include "Object.h"

// ======================================================================

ObjectList::ObjectList(int newMaxObjects, int newDepth, bool allowResize)
{
	depth  = newDepth;
	resize = allowResize;

	maxObjects = newMaxObjects;
	numObjects = 0;
	object     = NEW Object*[maxObjects];
	memset(object, 0, maxObjects * sizeof(*object));

	alterNum  = 0;
	alterSafe = NEW Object*[maxObjects];
}

// ----------------------------------------------------------------------

ObjectList::~ObjectList(void)
{
	maxObjects = 0;
	numObjects = 0;

	delete [] object;
	object = NULL;

	delete [] alterSafe;
	alterSafe = NULL;
}

// ----------------------------------------------------------------------

Object *ObjectList::getObject(int index)
{
	DEBUG_FATAL(index < 0 || index >= numObjects, ("ObjectList::getObject bad index %d [%d]", index, numObjects));
	return object[index];
}

// ----------------------------------------------------------------------

const Object *ObjectList::getObject(int index) const
{
	DEBUG_FATAL(index < 0 || index >= numObjects, ("ObjectList::getObject const bad index %d [%d]", index, numObjects));
	return object[index];
}

// ----------------------------------------------------------------------

bool ObjectList::findSlot(int uniqueId, int *slot) const
{
	bool found;
	int  low, high, mid;

	found = false;
	low   = 0;
	mid   = 0;
	high  = numObjects - 1;

	if (numObjects)
		while (!found && low <= high)
		{
			mid = (low + high) / 2;
			if (object[mid]->getUniqueId() == uniqueId)
				found = true;
			else
				if (object[mid]->getUniqueId() > uniqueId)
					high = mid - 1;
				else
 					low = mid + 1;
		}

	if (found)
		*slot = mid;
	else
		*slot = low;

	return found;
}

// ----------------------------------------------------------------------

const Object *ObjectList::findObject(int uniqueId) const
{
	int slot;

	if (findSlot(uniqueId, &slot))
		return object[slot];

	return NULL;
}

// ----------------------------------------------------------------------

const Object *ObjectList::findNextObject(const Object *obj, bool wrap) const
{
	int slot, result;

	// make sure we have objects
	if (!numObjects)
		return NULL;

	// null means give us the first object
	if (obj == NULL)
		return object[0];
	else
	{
		// find the current object
		result = findSlot(obj->getUniqueId(), &slot);
		DEBUG_FATAL(!result, ("ObjectList::findNextObject bad object"));
		
		// return the next object in the object list
		slot++;
		if (slot < numObjects)
			return object[slot];
		else
			if (wrap)
				return object[0];
			else
				return NULL;
	}	
}

// ----------------------------------------------------------------------

const Object *ObjectList::findPrevObject(const Object *obj, bool wrap) const
{
	int slot, result;

	// make sure we have objects
	if (!numObjects)
		return NULL;

	// null means give us the first object
	if (obj == NULL)
		return object[0];
	else
	{
		// find the current object
		result = findSlot(obj->getUniqueId(), &slot);
		DEBUG_FATAL(!result, ("ObjectList::findPrevObject bad object"));
		
		// return the next object in the object list
		slot--;
		if (slot >= 0)
			return object[slot];
		else
			if (wrap)
				return object[numObjects-1];
			else
				return NULL;
	}	
}

// ----------------------------------------------------------------------

void ObjectList::full(void)
{
	if (resize)
	{
		// grow twofold
		maxObjects *= 2;

		// create new lists
		Object** newObjectList = NEW Object* [maxObjects];
		memset (newObjectList, 0, sizeof (Object*) * maxObjects);

		Object** newAlterList  = NEW Object* [maxObjects];
		memset (newAlterList,  0, sizeof (Object*) * maxObjects);

		// copy old object list
		memcpy (newObjectList, object, sizeof (Object*) * numObjects);

		// reset
		delete [] object;
		object = newObjectList;

		delete [] alterSafe;
		alterSafe = newAlterList;
	}
	else
	{
		// -qq- resize the object list
		DEBUG_FATAL(1, ("ObjectList::full"));
	}
}

// ----------------------------------------------------------------------

void ObjectList::add(Object *obj)
{
	int i, j;

	// check for full case
	if (numObjects == maxObjects)
		full();

	// handle insertion as last element rapidly (this will be the common case)
	if (numObjects == 0 || object[numObjects-1]->getUniqueId() < obj->getUniqueId())
		object[numObjects++] = obj;
	else
	{
		bool result;

		// find sorted place to insert object
		result = findSlot(obj->getUniqueId(), &i);
		DEBUG_FATAL(result, ("ObjectList::add object already present"));

		// move objects up to make room
		for (j = numObjects; j > i; j--)
			object[j] = object[j-1];

		// insert the new object into the list
		object[i] = obj;
		numObjects++;
	}
}

// ----------------------------------------------------------------------

void ObjectList::remove(const Object *obj, bool del, bool removeFromAlterSafe)
{
	int found, slot;

	// search for the object
	found = findSlot(obj->getUniqueId(), &slot);
	if (found)
	{
		if (del)
		{
			// delete the object list copy to avoid an icky const_cast
			delete object[slot];
		}

		// copy all the elements down to preserve order
		for (numObjects--; slot < numObjects; slot++)
			object[slot] = object[slot+1];

		// whack the last object pointer for safety
		object[slot] = NULL;

		if (removeFromAlterSafe)
		{
			// search for the object in the alterSafe list.  can not do a binary search because
			// entries in this list are allowed to be NULL.
			for (slot = 0; slot < alterNum && alterSafe[slot] != obj; slot++)
				;

			// if it was found, just null it out
			if (slot < alterNum)
				alterSafe[slot] = NULL;
		}
	}
}

// ----------------------------------------------------------------------

void ObjectList::removeAll(bool del)
{
	Object *obj;

	while (numObjects)
	{
		obj = object[--numObjects];
		object[numObjects] = NULL;

		if (del)
			delete obj;
	}
}

// ----------------------------------------------------------------------

void ObjectList::render(void)
{
	int i;

	for (i = 0; i < numObjects; ++i)
		if (object[i] && object[i]->isActive())
			object[i]->render(true);
}

// ----------------------------------------------------------------------

void ObjectList::prepareToAlter(void)
{
	// copy the object list to a protected list for calling object alter routines
	alterNum = numObjects;
	if (alterNum)
		memcpy(alterSafe, object, alterNum * sizeof (*object));
}

// ----------------------------------------------------------------------

void ObjectList::alter(real time)
{
	Object *obj;
	int     i;

	// update using the copied list to protect object list integrity
	for (i = 0; i < alterNum; i++)
	{
		// cache the pointer
		obj = alterSafe[i];

		// test to see if the object is still alive
		if (obj)
		{
			// only the active alter OR the inactive alter should be called for an object
			int alterRemove;

			if (obj->isActive())
				alterRemove = obj->alter(time);
			else
				alterRemove = obj->inactiveAlter(time);

			if (alterRemove == 0)
			{
#if USE_OBJLIST_ALTER_REMOVE_AND_DELETE
				// check if the object has already been removed before doing another remove
				if (alterSafe[i])
				{
					alterSafe[i] =  NULL;
					remove(obj, false, false);
				}

				// delete the object
				delete obj;
#else
				DEBUG_FATAL( obj->isActive(), ("ObjectList::alter object alter returned false"));
				DEBUG_FATAL(!obj->isActive(), ("ObjectList::alter object inactiveAlter returned false"));
#endif
			}
		}
	}

	alterNum = 0;
}

// ----------------------------------------------------------------------

void ObjectList::addToCamera(Camera &camera) const
{
	int i;

	for (i = 0; i < numObjects; i++)
		object[i]->addToCamera(camera);
}

// ======================================================================
